/*
LINKFILE.C

C version of various pedigree and locus handling functions, swiped
for DOLINK so that locus input/output routines can be used,
but also used by MFLINK and other progs
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "linkfile.h"

#ifdef USEDCERROR
#include "dcerror.hpp"
#define error(s,t) error_func(__LINE__,__FILE__,s,t)
error_func(unsigned l,char *f,char *s1,char *s2)
{
dcerror(99,"Error on line %d of source file %s:\n %s%s",l,f,s1,s2);
return 0;
}
#else
error_func(unsigned l,char *f,char *s1,char *s2)
{
fprintf(stderr,"Error on line %d of source file %s:\n %s%s\n\n",l,f,s1,s2);
exit(1);
return 0;
}
#endif

int locus_def_output(FILE *fp,dc_locus_def *def,int sexlinked)
{
int i,j;
fprintf(fp,"%d %d # %s << type, number of alleles\n",def->type,def->nalls,def->name);
for (i=0;i<def->nalls;++i)
   fprintf(fp,"%f ",def->freq[i]);
fprintf(fp,"\n");
switch (def->type)
  {
  case DCL_AFF:
    fprintf(fp,"%d\n",def->info.aff.nliab);
    for (i=0;i<def->info.aff.nliab;++i)
      {
      fprintf(fp,"%f %f %f\n",
           def->info.aff.pen[i][0],
           def->info.aff.pen[i][1],
           def->info.aff.pen[i][2]);
      if (sexlinked)
          fprintf(fp,"%f %f\n",
           def->info.aff.male_pen[i][0],
           def->info.aff.male_pen[i][1]);
      }
    break;
  case DCL_BINARY:
    fprintf(fp,"%d\n",def->info.bin.nfact);
    for (i=0;i<def->nalls;++i)
      {
      for (j=0;j<def->info.bin.nfact;++j)
        fprintf(fp,"%c ",def->info.bin.fact[i][j]);
      fprintf(fp,"\n");
      }
    break;
  case DCL_QUAN:
    fprintf(fp,"%f %f",def->info.quan.mAA,&def->info.quan.maa);
    fprintf(fp,"%f",def->info.quan.dom);
    fprintf(fp,"%f %f",def->info.quan.var,def->info.quan.hetmult);
        break;
  case DCL_NUMBERED:
    break;
  }
if (def->isrisk) fprintf(fp,"%d\n",def->riskall);
fprintf(fp,"\n");
return 1;
}

int locus_def_input(fp,def,sexlinked,risklocus)
FILE *fp;
dc_locus_def *def;
int sexlinked,risklocus;
{
char line[1000],rest[1000];
int i,j;
def->name[0]='\0';
do {
   if (!fgets(line,1000,fp)) return 0;
   } while (sscanf(line,"%d %d # %s %[^\n]",&def->type,&def->nalls,def->name,rest)<2);
if (!fgets(line,1000,fp)) return 0;
def->isrisk=risklocus;
if ((def->type==DCL_AFF||def->type==DCL_QUAN) && def->nalls!=2)
  { error("Sorry, can only handle affection and quantitative loci with two alleles",""); return 0; }
if (def->nalls>MAXALL_LF)
  { error("Number of alleles exceeds MAXALL_LF at locus ",def->name); return 0; }
for (i=0;i<def->nalls;++i,strcpy(line,rest),*rest='\0')
   if (sscanf(line,"%f %[^\n]",&def->freq[i],rest)<1 ||
         def->freq[i]>1.0)  /* load floating point */
   { error("Problem reading allele frequencies",""); return 0; }
switch (def->type)
  {
  case DCL_AFF:
    if (!fgets(line,1000,fp) ||
          sscanf(line,"%d",&def->info.aff.nliab)!=1) return 0;
    for (i=0;i<def->info.aff.nliab;++i)
      {
      if (!fgets(line,1000,fp) ||
          sscanf(line,"%f %f %f",
           &def->info.aff.pen[i][0],
           &def->info.aff.pen[i][1],
           &def->info.aff.pen[i][2])!=3) return 0;
      if (sexlinked && (!fgets(line,1000,fp) ||
          sscanf(line,"%f %f",
           &def->info.aff.male_pen[i][0],
           &def->info.aff.male_pen[i][1])!=2)) return 0;
      }
    break;
  case DCL_BINARY:
    if (!fgets(line,1000,fp) ||
          sscanf(line,"%d",&def->info.bin.nfact)!=1) return 0;
    for (i=0;i<def->nalls;++i)
      {
      if (!fgets(line,1000,fp)) return 0;
      for (j=0;j<def->info.bin.nfact;++j,strcpy(line,rest))
        if (sscanf(line,"%c %[^\n]",&def->info.bin.fact[i][j],rest)<1) return 0;
      }
    break;
  case DCL_QUAN:
    if (!fgets(line,1000,fp) ||
          sscanf(line,"%f %f",&def->info.quan.mAA,&def->info.quan.maa)!=2) return 0;
    if (!fgets(line,1000,fp) ||
          sscanf(line,"%f",&def->info.quan.dom)!=1) return 0;
    if (!fgets(line,1000,fp) ||
          sscanf(line,"%f %f",&def->info.quan.var,&def->info.quan.hetmult)!=2) return 0;
        break;
  case DCL_NUMBERED:
    break;
  }
if (def->isrisk&&(!fgets(line,1000,fp)||sscanf(line,"%d",&def->riskall)!=1))
  return 0;
return 1;
}

read_person(FILE *fp,dc_person *ind,dc_locus_def *def,int nloci)
{
int i;
char line[2001],rest[2000],*ptr;
ind->nspouse=0;
ind->parents=NULL;
rest[0]='\0';
if (!fgets(line,2000,fp) || sscanf(line,"%s %d %d %d %d %d %d %d %d %[^\n]",
  &ind->fname,&ind->idnum,&ind->fidnum,&ind->midnum,
  &ind->chnum,&ind->fsibnum,&ind->msibnum,
  &ind->sex,&ind->isproband,rest)!=10) return 0;
ptr=rest;
for (i=0;i<nloci;++i)
  if (!locus_input(&ind->locus[i],&def[i],&ptr))
    {
    error("Could not read all locus data in line ",line);
    return 0;
    }
return 1;
}

write_person(FILE *fp,dc_person *ind,dc_locus_def *def,int nloci)
{
int i;
char line[2001],*ptr;
sprintf(line,"%8s %3d %3d %3d %3d %3d %3d %2d %2d",
  ind->fname,ind->idnum,ind->fidnum,ind->midnum,
  ind->chnum,ind->fsibnum,ind->msibnum,
  ind->sex,ind->isproband);
ptr=strchr(line,'\0');
for (i=0;i<nloci;++i)
  if (!locus_output(&ind->locus[i],&def[i],&ptr)) return 0;
fprintf(fp,"%s\n",line);
return 1;
}

write_pedigree(FILE *fp,dc_person *ped,dc_locus_def *def,int nloci,int count)
{
int i;
for (i=0;i<count;++i)
  if (!write_person(fp,&ped[i],def,nloci)) return 0;
return i;
}

read_pedigree(FILE *fp,dc_person *ped,dc_locus_def *def,int nloci)
{
static dc_person last_one;
int i;
if (last_one.idnum==0)
  if (!read_person(fp,&last_one,def,nloci)) return 0;
ped[0]=last_one;
i=1;
/* set last_one.idnum back to 0 if read fails */
/* gives possible unintended assignement warning, but
   assignment is intended */
while ((read_person(fp,&last_one,def,nloci)||(last_one.idnum=0)) &&
         !strcmp(ped[i-1].fname,last_one.fname))
        ped[i++]=last_one;
return i;
}

set_parents(dc_person *child,dc_person *ped,int count)
{
dc_mating *mating;
int i,j,k;
char buff[50];
for (i=0;i<count;++i)
  if (ped[i].idnum==child->fidnum) break;
if (i==count)
  {
  sprintf(buff,"%s_%d",child->fname,child->idnum);
  error("Father not found for ",buff);
  return 0;
  }
for (j=0;j<ped[i].nspouse;++i)
  if (ped[i].spouse[j]->mother->idnum==child->midnum) break;
if (j==ped[i].nspouse) /* mating not already found */
  {
  for (k=0;k<count;++k)
    if (ped[k].idnum==child->midnum) break;
  if (k==count)
    {
    sprintf(buff,"%s_%d",child->fname,child->idnum);
    error("Mother not found for ",buff);
    return 0;
    }
  mating=(dc_mating*)malloc(sizeof(dc_mating));
  if (mating==NULL) { error("No memory creating mating",""); return 0; }
  mating->nchild=0;
  mating->father=&ped[i];
  mating->mother=&ped[k];
  ped[i].spouse[ped[i].nspouse++]=mating;
  ped[k].spouse[ped[k].nspouse++]=mating;
  }
child->parents=ped[i].spouse[j];
child->parents->child[child->parents->nchild++]=child;
return 1;
}

free_matings(dc_person *ped,int count)
{
int i;
for (i=0;i<count;++i)
  if (ped[i].sex==1)
    while (ped[i].nspouse)
      free(ped[i].spouse[--ped[i].nspouse]);
return 1;
}

make_matings(dc_person *ped,int count)
{
int i;
for (i=0;i<count;++i)
  if (ped[i].fidnum!=0 && set_parents(&ped[i],ped,count)==0) return 0;
return 1;
}

int locus_input(dc_locus *locus,dc_locus_def *def,char **line)
{
int nread,n,i;
locus->type=def->type;
switch (def->type)
  {
  case DCL_AFF:
    if (def->info.aff.nliab==1)
      {
      if (sscanf(*line," %d%n",&locus->data.aff[0],&nread)!=1) return 0;
      }
    else
      if (sscanf(*line," %d %d%n",&locus->data.aff[0],
            &locus->data.aff[1],&nread)!=2) return 0;
    break;
  case DCL_QUAN:
    if (sscanf(*line," %f%n",&locus->data.qval,&nread)!=1) return 0;
    break;
  case DCL_BINARY:
    for (i=0,nread=0;i<def->info.bin.nfact;++i,nread+=n)
      if (sscanf(*line," %c%n",&locus->data.bin[i],&n)!=1) return 0;
    break;
  case DCL_NUMBERED:
      if (sscanf(*line," %d %d%n",&locus->data.all[0],
            &locus->data.all[1],&nread)!=2) return 0;
/* if one allele is 0 assume meant to be male with sex-linked
   marker (as in GAW 11 dataset) and set both alleles the same */
      if (locus->data.all[0]==0 && locus->data.all[1]!=0)
         locus->data.all[0]=locus->data.all[1];
      if (locus->data.all[0]!=0 && locus->data.all[1]==0)
         locus->data.all[1]=locus->data.all[0];
    break;
  }
*line+=nread;
return 1;
}

int locus_output(dc_locus *locus,dc_locus_def *def,char **line)
{
int nread,n,i;
switch (def->type)
  {
  case DCL_AFF:
    if (def->info.aff.nliab==1)
      sprintf(*line,"  %2d",locus->data.aff[0]);
    else
      sprintf(*line,"  %2d %-2d",locus->data.aff[0],
            locus->data.aff[1]);
    break;
  case DCL_QUAN:
    sprintf(*line,"  %8f",locus->data.qval);
  case DCL_BINARY:
    sprintf(*line,"  ");
    for (i=0;i<def->info.bin.nfact;++i)
      sprintf(*line," %c",locus->data.bin[i]);
    break;
  case DCL_NUMBERED:
      sprintf(*line,"  %3d %-3d",locus->data.all[0],
            locus->data.all[1]);
    break;
  }
*line=strchr(*line,'\0');
return 1;
}

read_data_file(FILE *fp,dc_locus_def *def,int *sexlinked,int *risklocus)
{
int nloci,i;
char line[1001];
if (!fgets(line,1000,fp) ||
      sscanf(line,"%d %d %d",&nloci,risklocus,sexlinked)!=3 ||
      !fgets(line,1000,fp) || !fgets(line,1000,fp))
        { error("Bad datafile",""); return 0; }
if (nloci>MAXLOCI_LF)
  { error("Number of loci exceeds MAXLOCI_LF",""); return 0; }
for (i=0;i<nloci;++i)
   {
   if (!locus_def_input(fp,&def[i],*sexlinked,*risklocus==i+1)) return 0;
   if (def[i].name[0]=='\0' || !(isalpha(def[i].name[0])||isdigit(def[i].name[0])))
     sprintf(def[i].name,"LOC%d",i+1);
   }
return nloci;
}

read_data_file_ordered(FILE *fp,dc_locus_def *def,int *sexlinked,int *risklocus,int *order,float *rf)
{
int nloci,i;
char line[2001],rest[2000];
if (!fgets(line,2000,fp) ||
      sscanf(line,"%d %d %d",&nloci,risklocus,sexlinked)!=3 ||
      !fgets(line,2000,fp) || !fgets(line,2000,fp))
        { error("Bad datafile",""); return 0; }
for (i=0;i<nloci;++i)
   if (sscanf(line,"%d %[^\n]",&order[i],rest)<1)
        { error("Not enough locus numbers in order line of datafile",""); return 0; }
   else strcpy(line,rest);
for (i=0;i<nloci;++i)
   {
   if (!locus_def_input(fp,&def[i],*sexlinked,*risklocus==i+1)) return 0;
   if (def[i].name[0]=='\0' || !(isalpha(def[i].name[0])||isdigit(def[i].name[0])))
     sprintf(def[i].name,"LOC%d",i+1);
   }
do {
   if (!fgets(line,2000,fp)) return nloci;
    /* not necessarily an error - calling program will have to check whether rfs read */
   } while (sscanf(line,"%f ",&rf[0])!=1);
if (rf[0]!=0) {error("Can only handle equal male:female recombination",""); return 0; }
do {
   if (!fgets(line,2000,fp)) return nloci;
    /* not necessarily an error - calling program will have to check whether rfs read */
   } while (sscanf(line,"%f ",&rf[0])!=1);
for (i=0;i<nloci-1;++i)
   if (sscanf(line,"%f %[^\n]",&rf[i],rest)<1)
        { error("Not enough recombination fractions in datafile",""); return 0; }
   else strcpy(line,rest);
return nloci;
}

write_data_file_ordered(FILE *fp,dc_locus_def *def,int sexlinked,int risklocus,int *order,float *rf,int nloci)
{
/* does not cope with risk locus */
int i;
fprintf(fp,"%d 0 %d 5\n0 0.0 0.0 0\n",nloci,sexlinked);
for (i=0;i<nloci;++i)
   fprintf(fp,"%2d%s",order[i],i==nloci-1?"\n\n":" ");
for (i=0;i<nloci;++i)
   locus_def_output(fp,&def[i],sexlinked);
fprintf(fp,"\n0 0\n");
for (i=0;i<nloci-1;++i)
   fprintf(fp,"%f%s",rf[i],i==nloci-2?"\n":" ");
return 1;
}
