/* PED2SDT.C */
/* Copyright Dave Curtis 1995 */
/* dcurtis@hgmp.mrc.ac.uk */
/* no warranty or liability of any kind is accepted, expressed or implied */

#define PROGNAME "PED2SDT"
#define VERSION "1.8"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "linkfile.h"

#define MAXSIB 30

char *datafile="datafile.dat";
char *pedfile="pedfile.dat";
char *sdtfile="sdtfile.dat";
int affloc=0,markloc=1;

int maxped2sdt()
{
printf("\nThe program sets the following maximum values:\n");
/*printf("%d pedigrees in the analysis \n",MAXPEDS);*/
printf("%d individuals in one pedigree \n",MAXIND_LF);
printf("%d affected or unaffected sibs \n",MAXSIB);
printf("%d loci \n",MAXLOCI_LF);
printf("%d alleles at any locus \n",MAXALL_LF);
return 0;
}

write_sdt(FILE *fo,dc_person *ped,int count,int sexlinked)
{
int i,j,c,a,naff,nunaff,aff[MAXSIB][2],unaff[MAXSIB][2];
dc_person sib;
for (i=0;i<count;++i)
  if (ped[i].sex==1 && ped[i].nspouse!=0)
    for (j=0;j<ped[i].nspouse;++j)
      {
      naff=nunaff=0;
      for (c=0;c<ped[i].spouse[j]->nchild;++c)
        {
        sib=*ped[i].spouse[j]->child[c];
        if (sib.locus[markloc].data.all[0]!=0)
          if (sib.locus[affloc].data.aff[0]==2)
            {
            for (a=0;a<2;++a)
              aff[naff][a]=sib.locus[markloc].data.all[a];
            ++naff;
            }
          else
            {
            for (a=0;a<2;++a)
              unaff[nunaff][a]=sib.locus[markloc].data.all[a];
            ++nunaff;
            }
        }
      if (naff!=0)
        {
        fprintf(fo,"%2d %2d %2d %2d %2d %2d ",naff,nunaff,
          ped[i].locus[markloc].data.all[0],
          ped[i].locus[markloc].data.all[1],
          ped[i].spouse[j]->mother->locus[markloc].data.all[0],
          ped[i].spouse[j]->mother->locus[markloc].data.all[1]);
        for (c=0;c<naff;++c)
          for (a=0;a<2;++a)
            fprintf(fo,"%2d ",aff[c][a]);
        for (c=0;c<nunaff;++c)
          for (a=0;a<2;++a)
            fprintf(fo,"%2d ",unaff[c][a]);
        fprintf(fo,"   << %s %d\n",ped[i].fname,ped[i].idnum);
        }
      }
return 1;
}


main(int argc,char *argv[])
{
FILE *pf,*df,*tf;
dc_locus_def def[MAXLOCI_LF];
dc_person *ped;
int nloci,count,sexlinked,risklocus;
printf("\n%s v%s\n",PROGNAME,VERSION);
maxped2sdt();
if (argc>1) pedfile=argv[1];
if (argc>2) datafile=argv[2];
if (argc>3) sdtfile=argv[3];
if (argc>4) { sscanf(argv[4],"%d",&affloc); --affloc; }
if (argc>5) { sscanf(argv[5],"%d",&markloc); --markloc; }
df=fopen(datafile,"r");
if (df==NULL) { error("Could not open file: ",datafile); exit(1); }
pf=fopen(pedfile,"r");
if (pf==NULL) { error("Could not open file: ",pedfile); exit(1); }
tf=fopen(sdtfile,"w");
if (tf==NULL) { error("Could not open file: ",sdtfile); exit(1); }
assert(pf && df && tf);
ped=(dc_person*)calloc(sizeof(dc_person),MAXIND_LF);
if (!ped) error("Out of memory allocating ped","");
if ((nloci=read_data_file(df,def,&sexlinked,&risklocus))==0) exit(1);
if (sexlinked) error("SDT does not handle sex-linked data yet","");
if (affloc<0 || markloc<0 || affloc>=nloci || markloc>=nloci ||
      def[affloc].type!=DCL_AFF || def[markloc].type!=DCL_NUMBERED)
  { error("Loci do not consist of an affection locus and a numbered locus",""); exit(5); }
fprintf(tf,"%d   << number of alleles\n",def[markloc].nalls);
while ((count=read_pedigree(pf,ped,def,nloci))!=0)
  {
  if (!make_matings(ped,count)) exit(2);
  if (!write_sdt(tf,ped,count,sexlinked)) exit(4);
  if (!free_matings(ped,count)) exit(3);
  }
free(ped);
fclose(df);
fclose(pf);
fclose(tf);
printf("\n%s written OK\n",sdtfile);
exit(0);
}
