#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dvector.hpp"
extern "C" {
#include "fisher.h"
}

#define PROGNAME "SDT"
#define VERSION "2.0"

#ifndef MAXSIB
#define MAXSIB 10
#endif
#ifndef MAXFAM
#define MAXFAM 10001
#endif
#ifndef MAXALL
#define MAXALL 40
#endif

enum {SDT_AND_TDT=0,ONLY_SDT,ONLY_TDT};
int SDT_OR_TDT=SDT_AND_TDT;

struct sdt_sub_t { int all[2]; int aff; };
struct nucfam_t { struct sdt_sub_t father, mother, sib[MAXSIB]; int naff,nunaff; }
  fam_sdt[MAXFAM],fam_tdt[MAXFAM];
dv2d Ssdt,Wsdt,Tsdt(1,1),Stdt,Wtdt,Ttdt(1,1),Sboth,Wboth,Tboth(1,1),Sm1,Wm1;

#define error(s,t) error_func(__LINE__,__FILE__,s,t)
char *format=" %8.0f";

void do_sdt(struct nucfam_t *fam,int nfam,int nall),do_tdt(struct nucfam_t *fam,int nfam,int nall);

int allsame_sigma_S=1,first_sigma_S=99;

int error_func(unsigned l,char *f,char *s1,char *s2)
{
fprintf(stderr,"Error on line %d of source file %s:\n %s%s\n\n",l,f,s1,s2);
exit(1);
return 0;
}

double chistat(double x,double df)
{
if (x==0.0) return 1.0;
else return 1.0-fisher(df,5000,x/df);
}

int maxsdt()
{
printf("\nThe program sets the following maximum values:\n");
/*printf("%d pedigrees in the analysis \n",MAXPEDS);*/
printf("%d maximum number of SDT or TDT families\n",MAXFAM);
printf("%d affected or unaffected sibs\n",MAXSIB);
printf("%d alleles at any locus \n",MAXALL);
return 0;
}

dv2d getT(dv2d &S,dv2d &W,int nall)
{
int rank,i,j,newi,newj;
for (rank=0,i=0;i<nall;++i)
  if (W[i][i]!=0) ++rank;
dv2d newW(rank,rank);
dv2d newS(rank,1);
for (i=0,newi=0;i<nall;++i)
  if (W[i][i]!=0)
    {
	newS[newi][0]=S[i][0];
    newW[newi][newi]=W[i][i];
	for (j=0,newj=0;j<i;++j)
	  if (W[j][j]!=0)
	    {
		newW[newi][newj]=W[i][j];
		newW[newj][newi]=W[j][i];
		++newj;
		}
    ++newi;
	}
dv2d T=newS.transpose()*newW.inv()*newS;
return T;
}

main(int argc, char *argv[])
{
FILE *fi,*fo;
int nfam_tdt,nfam_sdt,i,j,nall,use_tdt,nonzero,k,l,keptW;
char line[201],rest[201],keptline[201];
float p,df;
struct nucfam_t nuc;
printf("\n%s v%s\n",PROGNAME,VERSION);
maxsdt();
printf("\nIf you use this program, please cite:\n"
"Curtis D, Miller MB, Sham PC. Combining the sibling disequilibrium test and \n"
"transmission/disequilibrium test for multiallelic markers. Am J Hum Genet 1999 64 1785-6.\n\n");


if (argc>3 && argv[3][0]=='-')
  {
  if (argv[3][1]=='t') SDT_OR_TDT=ONLY_TDT;
  else if (argv[3][1]=='s') SDT_OR_TDT=ONLY_SDT;
  }
if (SDT_OR_TDT!=SDT_AND_TDT) printf("Will only use %cDT analysis\n",SDT_OR_TDT==ONLY_TDT?'T':'S');
fi=fopen(argv[1],"r");
if (!fi) error("Could not open file ",argv[1]);
nfam_sdt=nfam_tdt=0;
fgets(line,200,fi);
sscanf(line,"%d",&nall);
if (sscanf(line,"%d",&nall)!=1 || nall<=0)
  error("Bad number of alleles: ",line);
while (fgets(line,200,fi))
  {
  strcpy (keptline,line);
  i=sscanf(line,"%d %d %d %d %d %d %[^\n]",
    &nuc.naff,&nuc.nunaff,
    &nuc.father.all[0],&nuc.father.all[1],
    &nuc.mother.all[0],&nuc.mother.all[1],rest);
  if (i<2) continue; /* skip blank lines */
  else if (i!=7)
      error("Bad data format: ",keptline);
  strcpy(line,rest);
  for (i=0;i<nuc.naff;++i)
    {
    if (sscanf(line,"%d %d %[^\n]",&nuc.sib[i].all[0],&nuc.sib[i].all[1],rest)<2)
      error("Bad data format: ",keptline);
    strcpy(line,rest);
    if (nuc.sib[i].all[0]==0)
      {
      --i;
      --nuc.naff;
      // skip sibs which have not been genotyped
      }
    }
  if (nuc.naff==0) error("Must be at least one affected sib genotyped: ",keptline);
  for (i=nuc.naff;i<nuc.naff+nuc.nunaff;++i)
    {
    if (sscanf(line,"%d %d %[^\n]",&nuc.sib[i].all[0],&nuc.sib[i].all[1],rest)<2)
      error("Bad data format: ",keptline);
    strcpy(line,rest);
    if (nuc.sib[i].all[0]==0)
      {
      --i;
      --nuc.nunaff;
      // skip sibs which have not been genotyped
      }
    }
  use_tdt=((nuc.father.all[0]!=0 && nuc.mother.all[0]!=0) ||
    ((nuc.father.all[0]!=nuc.father.all[1]) ||
    (nuc.mother.all[0]!=nuc.mother.all[1]))&&nall>2);
  if (SDT_OR_TDT!=SDT_AND_TDT) use_tdt=SDT_OR_TDT==ONLY_TDT;
  if (use_tdt)
    {
    if (nfam_tdt>=MAXFAM) error("Too many TDT families: ",keptline);
    else fam_tdt[nfam_tdt++]=nuc;
    }
  else
    {
    if (nfam_sdt>=MAXFAM) error("Too many SDT families: ",keptline);
    else fam_sdt[nfam_sdt++]=nuc;
    }
  }
fo=fopen(argv[2],"w");
if (!fo) error("Could not open file ",argv[2]);
if (SDT_OR_TDT==ONLY_TDT)
  fprintf(fo,"Only TDT analysis to be performed\n\n");
else if (SDT_OR_TDT!=ONLY_TDT && nfam_sdt>0)
  {
  do_sdt(fam_sdt,nfam_sdt,nall);
  fprintf(fo,"Ssdt:\n (difference between number of sibships in which allele is commoner in cases or controls):\n");
  Ssdt.transpose().fprintf(fo,format);
  fprintf(fo,"Total informative sibships for each allele:\n");
  for (i=0;i<nall;++i) fprintf(fo,format,Wsdt[i][i]); // I think this is right
  fprintf(fo,"\nSingle allele p value, uncorrected for multiple alleles:\n");
  for (i=0;i<nall;++i) fprintf(fo," %8.6f",chistat(Wsdt[i][i]?Ssdt[i][0]*Ssdt[i][0]/Wsdt[i][i]:0.0,1.0));
  fprintf(fo,"\n\nWsdt:\n");
  Wsdt.fprintf(fo,format);
  if (allsame_sigma_S) 
    {
	for (l=0;l<nall;++l)
	  {
	  if (Wsdt[l][l]!=0)
	    {
		keptW=Wsdt[l][l];
		Wsdt[l][l]=0; // trick to force matrix reduction
        Tsdt=getT(Ssdt,Wsdt,nall);
        for (i=0,df=0;i<nall;++i) df+=(Wsdt[i][i]!=0);
        p=chistat(Tsdt[0][0],df);
        fprintf(fo,"Tsdt = %6.3f, %.0f df, p = %8.6f\n\n",Tsdt[0][0],df,p);
        Wsdt[l][l]=keptW;
		break; // omit this line to check that always get same answer 
		       // regardless of which allele is dropped
        }
	  }
	}
  else
    {
    if (nall==2) Wsdt[1][1]=0; // trick to force matrix reduction
    Tsdt=getT(Ssdt,Wsdt,nall);
    for (i=0,df=0;i<nall;++i) df+=(Wsdt[i][i]!=0);
    p=chistat(Tsdt[0][0],df);
    fprintf(fo,"Tsdt = %6.3f, %.0f df, p = %8.6f\n\n",Tsdt[0][0],df,p);
	}
  }
else if (SDT_OR_TDT!=ONLY_TDT && nfam_sdt<=0)
  fprintf(fo,"No SDT families identified for analysis\n\n");
if (SDT_OR_TDT==ONLY_SDT)
  fprintf(fo,"Only SDT analysis to be performed\n\n");
else if (SDT_OR_TDT!=ONLY_SDT && nfam_tdt>0)
  {
  do_tdt(fam_tdt,nfam_tdt,nall);
  fprintf(fo,"Stdt:\n (difference between number of parents preferentially transmitting allele to cases or controls):\n");
  Stdt.transpose().fprintf(fo,format);
  fprintf(fo,"Total informative parents for each allele:\n");
  for (i=0;i<nall;++i) fprintf(fo,format,Wtdt[i][i]); // I think this is right
  fprintf(fo,"\nSingle allele p value, uncorrected for multiple alleles:\n");
  for (i=0;i<nall;++i) fprintf(fo," %8.6f",chistat(Wtdt[i][i]?Stdt[i][0]*Stdt[i][0]/Wtdt[i][i]:0.0,1.0));
  fprintf(fo,"\n\nWtdt:\n");
  Wtdt.fprintf(fo,format);
  if (1) // always have to do this for TDT analysis
    {
	for (l=0;l<nall;++l)
	  {
	  if (Wtdt[l][l]!=0)
	    {
		keptW=Wtdt[l][l];
		Wtdt[l][l]=0; // trick to force matrix reduction
        Ttdt=getT(Stdt,Wtdt,nall);
        for (i=0,df=0;i<nall;++i) df+=(Wtdt[i][i]!=0);
        p=chistat(Ttdt[0][0],df);
        fprintf(fo,"Ttdt = %6.3f, %.0f df, p = %8.6f\n\n",Ttdt[0][0],df,p);
        Wtdt[l][l]=keptW;
		break;
        }
	  }
	}
  }
else if (SDT_OR_TDT!=ONLY_SDT && nfam_tdt<=0)
  fprintf(fo,"No TDT families identified for analysis\n\n");
if (SDT_OR_TDT==SDT_AND_TDT && nfam_sdt>0 && nfam_tdt>0)
  {
  Sboth=Ssdt+Stdt;
  Wboth=Wsdt+Wtdt;
  Tboth=getT(Sboth,Wboth,nall);
  fprintf(fo,"Sboth:\n");
  Sboth.transpose().fprintf(fo,format);
  fprintf(fo,"Total informative observations for each allele:\n");
  for (i=0;i<nall;++i) fprintf(fo,format,Wboth[i][i]); // I think this is right
  fprintf(fo,"\nSingle allele p value, uncorrected for multiple alleles:\n");
  for (i=0;i<nall;++i) fprintf(fo," %8.6f",chistat(Wboth[i][i]?Sboth[i][0]*Sboth[i][0]/Wboth[i][i]:0.0,1.0));
  fprintf(fo,"\n\nWboth:\n");
  Wboth.fprintf(fo,format);
  if (allsame_sigma_S) 
    {
	for (l=0;l<nall;++l)
	  {
	  if (Wboth[l][l]!=0)
	    {
		keptW=Wboth[l][l];
		Wboth[l][l]=0; // trick to force matrix reduction
        Tboth=getT(Sboth,Wboth,nall);
        for (i=0,df=0;i<nall;++i) df+=(Wboth[i][i]!=0);
        p=chistat(Tboth[0][0],df);
        fprintf(fo,"Tboth = %6.3f, %.0f df, p = %8.6f\n\n",Tboth[0][0],df,p);
        Wboth[l][l]=keptW;
		break; // omit this line to check that always get same answer 
		       // regardless of which allele is dropped
        }
	  }
	}
  else
    {
    if (nall==2) Wboth[1][1]=0; // trick to force matrix reduction
    Tboth=getT(Sboth,Wboth,nall);
    for (i=0,df=0;i<nall;++i) df+=(Wboth[i][i]!=0);
    p=chistat(Tboth[0][0],df);
    fprintf(fo,"Tboth = %6.3f, %.0f df, p = %8.6f\n\n",Tboth[0][0],df,p);
	}
  }
return 0;
}

void do_sdt(struct nucfam_t *fam_sdt,int nfam_sdt,int nall)
{
int i,j,k,nonzero;
float freq[2][MAXALL];
int s[MAXALL],sigma_S;
Ssdt.resize(nall,1); Wsdt.resize(nall,nall);
Sm1.resize(nall-1,1); Wm1.resize(nall-1,nall-1);
for (j=0;j<nall;++j)
    {
    Ssdt[j][0]=0;
    for (k=0;k<nall;++k)
      Wsdt[j][k]=0;
    }
for (i=0;i<nfam_sdt;++i)
  {
  if (fam_sdt[i].nunaff==0) continue;
  for (j=0;j<nall;++j) freq[0][j]=freq[1][j]=0;
  for (j=0;j<fam_sdt[i].naff;++j)
    {
    ++freq[0][fam_sdt[i].sib[j].all[0]-1];
    ++freq[0][fam_sdt[i].sib[j].all[1]-1];
    }
  for (j=fam_sdt[i].naff;j<fam_sdt[i].naff+fam_sdt[i].nunaff;++j)
    {
    ++freq[1][fam_sdt[i].sib[j].all[0]-1];
    ++freq[1][fam_sdt[i].sib[j].all[1]-1];
    }
  for (j=0;j<nall;++j)
    {
    freq[0][j]/=fam_sdt[i].naff;
    freq[1][j]/=fam_sdt[i].nunaff;
    s[j]=freq[0][j]>freq[1][j]?1:freq[0][j]<freq[1][j]?-1:0;
    }
  for (j=0,sigma_S=0;j<nall;++j)
    {
	sigma_S+=s[j];
    Ssdt[j][0]+=s[j];
    for (k=0;k<nall;++k)
      Wsdt[j][k]+=s[j]*s[k];
    }
  if (first_sigma_S==99)
    first_sigma_S=sigma_S;
  else if (sigma_S!=first_sigma_S)
    allsame_sigma_S=0;
  }
}

static int swap_temp;
#define INTSWAP(m,n) (swap_temp=m,m=n,n=swap_temp)
enum {BOTH=0,PATERNAL,MATERNAL};
int which=BOTH;

void do_tdt(struct nucfam_t *fam_tdt,int nfam_tdt,int nall)
{
int i,j,k,c,nonzero;
int loc[2],floc[2],mloc[2];
int f_passed[MAXALL],f_not_passed[MAXALL],m_passed[MAXALL],m_not_passed[MAXALL];
int s[MAXALL];
Stdt.resize(nall,1);Wtdt.resize(nall,nall);
Sm1.resize(nall-1,1);Wm1.resize(nall-1,nall-1);
for (j=0;j<nall;++j)
    {
    Stdt[j][0]=0;
    for (k=0;k<nall;++k)
      Wtdt[j][k]=0;
    }
for (i=0;i<nfam_tdt;++i)
  {
  for (j=0;j<nall;++j)
    f_passed[j]=f_not_passed[j]=m_passed[j]=m_not_passed[j]=0;
  for (j=0;j<2;++j)
    {
    floc[j]=fam_tdt[i].father.all[j];
    mloc[j]=fam_tdt[i].mother.all[j];
    }
  if (floc[0]>floc[1]) INTSWAP(floc[0],floc[1]);
  if (mloc[0]>mloc[1]) INTSWAP(mloc[0],mloc[1]);
  for (c=0;c<fam_tdt[i].naff;++c)
    {
    for (j=0;j<2;++j)
      loc[j]=fam_tdt[i].sib[c].all[j];
    if (loc[0]==0) continue; // this should not happen as only genotyped sibs are included
    if ((floc[0]==0 || mloc[0]==0) && loc[0]==loc[1]) continue;
    // do not allow either parent unknown if child is homozygous
    if (loc[0]>loc[1]) INTSWAP(loc[0],loc[1]);
    if (loc[0]!=floc[0] || loc[1]!=floc[1])
      {
      if (floc[0]!=0)
        for (k=0;k<2;++k)
          for (j=0;j<2;++j)
            if (floc[k]==loc[j])
              {
              if (which!=MATERNAL)
                {
                ++f_passed[floc[k]-1];
                ++f_not_passed[floc[(k+1)%2]-1];
                }
              loc[j]=0;
              goto done_father;
              }
      done_father:
      if (mloc[0]!=0 && (loc[0]!=mloc[0] || loc[1]!=mloc[1]))
        for (k=0;k<2;++k)
          for (j=0;j<2;++j)
            if (mloc[k]==loc[j])
              {
              if (which!=PATERNAL)
                {
                ++m_passed[mloc[k]-1];
                ++m_not_passed[mloc[(k+1)%2]-1];
                }
              loc[j]=0;
              goto done_mother;
              }
      done_mother: ;
      }
    else /* same as father, just do mother first */
      if (loc[0]!=mloc[0] || loc[1]!=mloc[1])
        /* if same as mother too, one is passed and one isn't */
      {
      if (mloc[0]!=0)
        {
        for (k=0;k<2;++k)
          for (j=0;j<2;++j)
            if (mloc[k]==loc[j])
              {
              if (which!=PATERNAL)
                {
                ++m_passed[mloc[k]-1];
                ++m_not_passed[mloc[(k+1)%2]-1];
                }
              loc[j]=0;
              goto done_mother2;
              }
        done_mother2:
        for (k=0;k<2;++k)
          for (j=0;j<2;++j)
            if (floc[k]==loc[j])
              {
              if (which!=MATERNAL)
                {
                ++f_passed[floc[k]-1];
                ++f_not_passed[floc[(k+1)%2]-1];
                }
              loc[j]=0;
              goto done_father2;
              }
        done_father2: ;
        }
      }
      else if (which==BOTH)
      /* three genotypes identical */
	  /* we do not really know which is which */
        {
        ++f_passed[loc[0]-1];
        ++f_not_passed[loc[1]-1];
        ++m_passed[loc[1]-1];
        ++m_not_passed[loc[0]-1];
        }
    }
  for (j=0;j<nall;++j)
    s[j]=f_passed[j]>f_not_passed[j]?1:f_passed[j]<f_not_passed[j]?-1:0;
  for (j=0;j<nall;++j)
    {
    Stdt[j][0]+=s[j];
    for (k=0;k<nall;++k)
      Wtdt[j][k]+=s[j]*s[k];
    }
  for (j=0;j<nall;++j)
    s[j]=m_passed[j]>m_not_passed[j]?1:m_passed[j]<m_not_passed[j]?-1:0;
  for (j=0;j<nall;++j)
    {
    Stdt[j][0]+=s[j];
    for (k=0;k<nall;++k)
      Wtdt[j][k]+=s[j]*s[k];
    }
  }
}

