#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "dvector.hpp"
extern "C" {
float uran3( int *);
double snrm2( int, double *);
}

#define MAX(a,b)  ( (a) > (b) ? (a) : (b) )
#define MIN(a,b)  ( (a) < (b) ? (a) : (b) )

main(int argc, char *argv[])
{
	int		rows, cols, num_mat, trial, i, m, n, job,
	        info, seed, p;
	double	udiff, vdiff, sval_sum, frob_norm, frob_err,
			cpu_time, sys_time, fac, udiff_min, udiff_max, udiff_sum,
			vdiff_min, vdiff_max, vdiff_sum, frob_min, frob_max, frob_sum,
			stemp, recnstrct_norm, recnstrct_err, recnstrct_min,
			recnstrct_max, recnstrct_sum;
	double	ctemp, *e;

if (argc<4)
  {
  printf("Usage: testsvd numrows numcols numtrials\n");
  return 1;
  }
rows=atoi(argv[1]);
cols=atoi(argv[2]);
num_mat=atoi(argv[3]);

	/* allocate matrices and vectors */
	dv2d x(rows,cols);
	dsvd S(rows,cols);
//	doublevec s(MIN(cols,rows+1)),e(cols),work(rows);
    doublevec work(MAX(rows,cols));
    doublevec s(MIN(cols,rows+1));

	seed = -69;
	udiff_min = vdiff_min = frob_min = recnstrct_min = 1.e+20;
	udiff_max = vdiff_max = frob_max = recnstrct_max = -1.e+20;
	udiff_sum = vdiff_sum = frob_sum = recnstrct_sum = 0.0;

	for ( trial = 0;  trial < num_mat;  trial++ ) {

		/* fill matrix with uniform random numbers
         * of randomly chosen amplitude */
		fac = pow( 10.0, uran3( &seed ) *  10. - 5.0 );
		for ( n = 0;  n < cols;  n++ ) {
			for ( m = 0;  m < rows;  m++ ) {
				x[m][n] = fac * (uran3(&seed) - 0.5);
			}
		}

		/* compute SVD */

		if (S.dcmp(x)==0) printf("csvdc info was non-zero\n");
		S.setU(); // remove these later
	    S.setV();
#if 0
S.U.print("%6.3f ");
S.W.print("%6.3f ");
S.V.print("%6.3f ");
#endif
		for ( i = 0;  i < MIN(rows,cols);  i++ )
		 s[i]=S.W[i][i];

		udiff = 0.0;
// in the C routine these were indexed up to rows, because
// csvdc makes a rows x rows matrix for u with joba=1
		for ( n = 0;  n < MIN(rows,cols);  n++ ) {
			for ( m = 0;  m < MIN(rows,cols);  m++ ) {
			    for (i=0,ctemp=0;i<rows;++i)
			      ctemp+=S.UT[n][i]*S.UT[m][i];
				if ( n == m )
					ctemp -= 1.0;
				udiff += fabs( ctemp );
			}
		}
		udiff_min = MIN( udiff, udiff_min );;
		udiff_max = MAX( udiff, udiff_max );;
		udiff_sum += udiff;

		/* check orthonormality of v vectors */
		vdiff = 0.0;
		for ( n = 0;  n < cols;  n++ ) {
			for ( m = 0;  m < cols;  m++ ) {
			    for (i=0,ctemp=0;i<cols;++i)
			      ctemp+=S.VT[n][i]*S.VT[m][i];
				if ( n == m )
					ctemp -= 1.0;
				vdiff += fabs( ctemp );
			}
		}
		vdiff_min = MIN( vdiff, vdiff_min );;
		vdiff_max = MAX( vdiff, vdiff_max );;
		vdiff_sum += vdiff;

		/* check non-negativity of singular values */
		for ( n = 0;  n < MIN(rows,cols);  n++ ) {
			if ( s[n] < 0.0 ) {
				printf("Negative singular value # %d = (%e)\n",
					n, s[n]);
			}
		}

		/* check sum of squared singular values against Frobenius norm of x */
		sval_sum = 0.0;
		for ( n = 0;  n < MIN(rows,cols);  n++ )
			sval_sum += s[n] * s[n];
		frob_norm = 0.0;
		for ( n = 0;  n < cols;  n++ ) {
		    for (i=0;i<rows;++i) work[i]=x[i][n];
            stemp = snrm2( rows, work.vec() );
			frob_norm += stemp * stemp;
        }
		frob_err = fabs(sval_sum - frob_norm) / frob_norm;
		if ( frob_err > 1.e-4 )
			printf("Frobenius error = %e\n", frob_err);
		frob_min = MIN( frob_err, frob_min );
		frob_max = MAX( frob_err, frob_max );
		frob_sum += frob_err;

		x=x-S.U*S.W*S.VT;
		recnstrct_norm = 0.0;
		for ( n = 0;  n < cols;  n++ ) {
		    for (i=0;i<rows;++i) work[i]=x[i][n];
            stemp = snrm2( rows, work.vec() );
			recnstrct_norm += stemp * stemp;
        }
		recnstrct_err = recnstrct_norm / frob_norm;
		if ( recnstrct_err > 1.e-4 )
			printf("Reconstruction error = %e\n", recnstrct_err);
		recnstrct_min = MIN( recnstrct_err, recnstrct_min );
		recnstrct_max = MAX( recnstrct_err, recnstrct_max );
		recnstrct_sum += recnstrct_err;


	}

	printf("Orthonormality of u:\n");
	printf("       min_err = %e\n", udiff_min);
	printf("       max_err = %e\n", udiff_max);
	printf("       avg_err = %e\n", udiff_sum / num_mat);
	printf("Orthonormality of v:\n");
	printf("       min_err = %e\n", vdiff_min);
	printf("       max_err = %e\n", vdiff_max);
	printf("       avg_err = %e\n", vdiff_sum / num_mat);
	printf("Frobenius norm vs. (singular values)^2:\n");
	printf("       min_err = %e\n", frob_min);
	printf("       max_err = %e\n", frob_max);
	printf("       avg_err = %e\n", frob_sum / num_mat);
	printf("Reconstruction error of x - u*sigma*v:\n");
	printf("       min_err = %e\n", sqrt(recnstrct_min));
	printf("       max_err = %e\n", sqrt(recnstrct_max));
	printf("       avg_err = %e\n", sqrt(recnstrct_sum / num_mat));

    return 0;
}
